package core

import (
	"fmt"
	"math/big"
	"witness/tree"

	"github.com/ethereum/go-ethereum/common"
	"github.com/ethereum/go-ethereum/crypto"
	log "github.com/sirupsen/logrus"
)

func (w *Witness) GetPendingWorkload(address common.Address) (workload, globalWorkload uint64) {
	wl, err := w.q.GetPendingWorkload(w.todayTimestamp(), address.Hex())
	if err != nil {
		log.WithError(err).Error("failed to get pending workload")
		return
	}
	log.WithField("workload", wl).Debug("quest get pending workload")
	return wl, w.pendingWorkload
}

func (w *Witness) GetMerkleProof(address common.Address, date string) (balance string, proofs []common.Hash) {
	if date == "" {
		date = w.date
	}
	w.Lock()
	cacheTree, ok := w.mtTreeCache[date]
	w.Unlock()
	if !ok {
		if ok = w.LoadMerkleTree(date); !ok {
			log.WithFields(log.Fields{
				"date": date,
			}).Error("load merkle proof empty")
			return "0", nil
		}
	}
	w.Lock()
	cacheTree = w.mtTreeCache[date]
	w.Unlock()
	dateStateRootKey := fmt.Sprintf("sroot:%s", date)
	dateStateRoot, err := w.lvdb.Get([]byte(dateStateRootKey))
	if err != nil {
		log.WithFields(log.Fields{
			"key": dateStateRootKey,
			"err": err.Error(),
		}).Error("failed to get state root")
		return "0", nil
	}
	var sdb *StateDB
	if date == w.date {
		sdb = w.state
	} else {
		sdb, err = NewStateDB(w.lvdb, common.BytesToHash(dateStateRoot))
		if err != nil {
			log.WithError(err).Error("failed to create state db")
			return "0", nil
		}
	}
	object := sdb.GetMinerObject(address)
	if object == nil {
		return "0", nil
	}
	bigBalance, _ := new(big.Int).SetString(object.Balance, 10)
	payload := append(common.HexToAddress(object.Miner).Bytes(), common.LeftPadBytes(bigBalance.Bytes(), 32)...)
	leaf := crypto.Keccak256Hash(payload)
	proofs, err = cacheTree.GetProof(leaf)
	if err != nil {
		log.WithError(err).Error("failed to get merkle proof")
		return "0", make([]common.Hash, 0)
	}
	return object.Balance, proofs
}

func (w *Witness) GetDailyMerkleNodes(date string, depth int, rootHash common.Hash) (nodes [][]common.Hash) {
	if date == "" {
		date = w.date
	}
	w.Lock()
	cacheTree, ok := w.mtTreeCache[date]
	w.Unlock()
	if !ok {
		if ok = w.LoadMerkleTree(date); !ok {
			log.WithFields(log.Fields{
				"date": date,
			}).Error("load merkle proof empty")
			return nil
		}
	}
	w.Lock()
	cacheTree = w.mtTreeCache[date]
	w.Unlock()

	rootNode := cacheTree.GetRootNode()
	if rootHash.Hex() == (common.Hash{}).Hex() {
		rootNode = cacheTree.GetRootNode()
	} else if rootHash.Hex() != rootNode.Hash.Hex() {
		rootNode = cacheTree.FindMerkleNode(rootHash)
	}
	return tree.MerkleTreeTraversal(rootNode, depth)
}

func (w *Witness) GetDailyMerkleSumNodes(date string, depth int, rootHash common.Hash) (nodesHash [][]common.Hash, nodesVal [][]string) {
	if date == "" {
		date = w.date
	}
	w.Lock()
	cacheTree, ok := w.mstTreeCache[date]
	w.Unlock()
	if !ok {
		if ok = w.LoadMerkleSumTree(date); !ok {
			log.WithFields(log.Fields{
				"date": date,
			}).Error("load merkle sum proof empty")
			return nil, nil
		}
	}
	w.Lock()
	cacheTree = w.mstTreeCache[date]
	w.Unlock()

	rootNode := cacheTree.GetRoot()
	if rootHash.Hex() == (common.Hash{}).Hex() {
		rootNode = cacheTree.GetRoot()
	} else if rootHash.Hex() != rootNode.Value.Hash.Hex() {
		rootNode = cacheTree.FindMerkleSumNode(rootHash)
	}
	vlss := tree.MerkleSumTreeTraversal(rootNode, depth)
	for i := 0; i < len(vlss); i++ {
		_nodesHash := make([]common.Hash, 0)
		_nodesVal := make([]string, 0)
		for j := 0; j < len(vlss[i]); j++ {
			_nodesHash = append(_nodesHash, vlss[i][j].Hash)
			_nodesVal = append(_nodesVal, vlss[i][j].BigValue.String())
		}
		nodesHash = append(nodesHash, _nodesHash)
		nodesVal = append(nodesVal, _nodesVal)
	}
	return nodesHash, nodesVal
}
