package core

import (
	"bytes"
	"fmt"
	"math/big"
	"sort"
	"time"
	"witness/tree"
	"witness/util"

	"github.com/ethereum/go-ethereum/common"
	"github.com/ethereum/go-ethereum/crypto"
	witnessv1 "github.com/odysseus/odysseus-protocol/gen/proto/go/witness/v1"
	log "github.com/sirupsen/logrus"
	"github.com/syndtr/goleveldb/leveldb/errors"
)

// CommitMST commit workload of per day
func (w *Witness) CommitMST(proofMap map[common.Address]*witnessv1.ValidatedProof) (root common.Hash, sum *big.Int, err error) {
	if len(proofMap) == 0 {
		return common.Hash{}, big.NewInt(0), nil
	}
	rawKeys := make([]common.Address, 0)
	keys := make([][]byte, len(proofMap))
	dbKey := make([]byte, 0)
	dbVal := make([]byte, 0)
	vals := make([]*big.Int, len(proofMap))
	for key := range proofMap {
		rawKeys = append(rawKeys, key)
	}
	sort.Sort(util.AddressSlice(rawKeys))
	for i, key := range rawKeys {
		dbKey = append(dbKey, rawKeys[i][:]...)
		keys[i] = rawKeys[i].Bytes()
		vals[i] = new(big.Int).SetUint64(proofMap[key].Workload)
		dbVal = append(append(dbVal, vals[i].Bytes()...), []byte(":")...)
	}

	st := time.Now()

	mstTree := tree.NewMerkleSumTree(keys, vals)

	rootNode := mstTree.GetRoot()

	err = w.lvdb.Put([]byte(fmt.Sprintf("mstroot:%s", w.date)), root.Bytes())
	if err != nil {
		log.Error(err)
		return
	}
	err = w.lvdb.Put([]byte(fmt.Sprintf("mstsum:%s", w.date)), sum.Bytes())
	if err != nil {
		log.Error(err)
		return
	}

	err = w.lvdb.Put([]byte(fmt.Sprintf("mstk:%s", w.date)), dbKey)
	if err != nil {
		return
	}
	err = w.lvdb.Put([]byte(fmt.Sprintf("mstv:%s", w.date)), dbVal)
	if err != nil {
		return
	}
	log.WithFields(log.Fields{
		"root": root.Hex(),
		"sum":  sum.String(),
		"cost": time.Since(st).String(),
	}).Info("commit MST root")

	return rootNode.Value.Hash, rootNode.Value.BigValue, nil
}

// CommitMT commit all workload
func (w *Witness) CommitMT(objects []*witnessv1.MinerObject) (root common.Hash, err error) {
	if len(objects) == 0 {
		return common.Hash{}, nil
	}
	merkleProofs := make(tree.Proofs, 0)
	dbProofs := make([]byte, 0)
	for _, object := range objects {
		bigBalance, _ := new(big.Int).SetString(object.Balance, 10)
		payload := append(common.HexToAddress(object.Miner).Bytes(), common.LeftPadBytes(bigBalance.Bytes(), 32)...)
		_proof := crypto.Keccak256Hash(payload)
		merkleProofs = append(merkleProofs, _proof)
		dbProofs = append(dbProofs, _proof[:]...)
	}

	mtTree, err := tree.NewMerkleTree(merkleProofs)
	if err != nil {
		return
	}
	root = mtTree.GetRoot()

	st := time.Now()
	err = w.lvdb.Put([]byte(fmt.Sprintf("mtroot:%s", w.date)), root.Bytes())
	if err != nil {
		return
	}
	log.WithFields(log.Fields{
		"k": fmt.Sprintf("mtroot:%s", w.date),
		"v": root.String(),
	}).Debug()

	err = w.lvdb.Put([]byte(fmt.Sprintf("mtk:%s", w.date)), dbProofs)
	if err != nil {
		return
	}
	log.WithFields(log.Fields{
		"k":        fmt.Sprintf("mtk:%s", w.date),
		"v_length": len(dbProofs),
	}).Debug()
	w.mtTreeCache[w.date] = mtTree

	log.WithFields(log.Fields{
		"root": root.Hex(),
		"cost": time.Since(st).String(),
	}).Info("commit MT root")

	return
}

func (w *Witness) LoadMerkleTree(date string) (ok bool) {
	if date == "" {
		return false
	}
	merkleTreeKey := fmt.Sprintf("mtk:%s", date)
	data, err := w.lvdb.Get([]byte(merkleTreeKey))
	if err != nil {
		if err == errors.ErrNotFound {
			return
		}
		log.WithError(err).Error("failed to load merkle proof")
		return
	}

	log.WithFields(log.Fields{
		"key":    merkleTreeKey,
		"length": len(data),
	}).Info("diskdb load merkle proof")

	proofs := make([]common.Hash, len(data)/32)
	for i := 0; i < len(data)/32; i++ {
		copy(proofs[i][:], data[i*32:(i+1)*32])
	}

	mTree, err := tree.NewMerkleTree(proofs)
	if err != nil {
		log.WithError(err).Error("failed to load merkle proof")
		return
	}
	w.Lock()
	w.mtTreeCache[date] = mTree
	w.Unlock()
	log.WithFields(log.Fields{"date": date, "root": mTree.GetRoot().Hex()}).Info("load merkle tree")
	return true
}

func (w *Witness) LoadMerkleSumTree(date string) (ok bool) {
	if date == "" {
		return false
	}
	merkleSumTreeKey := fmt.Sprintf("mstk:%s", date)
	keyData, err := w.lvdb.Get([]byte(merkleSumTreeKey))
	if err != nil {
		if err == errors.ErrNotFound {
			return
		}
		log.WithError(err).Error("failed to load merkle sum proof k")
		return
	}
	log.WithFields(log.Fields{
		"key":    merkleSumTreeKey,
		"length": len(keyData),
	}).Info("diskdb load merkle sum proof key")

	// data = addr1:addr2
	datas := make([][]byte, len(keyData)/20)
	for i := 0; i < len(keyData)/20; i++ {
		copy(datas[i], keyData[i*20:(i+1)*20])
	}

	merkleSumTreeVal := fmt.Sprintf("mstv:%s", date)
	valData, err := w.lvdb.Get([]byte(merkleSumTreeVal))
	if err != nil {
		if err == errors.ErrNotFound {
			return
		}
		log.WithError(err).Error("failed to load merkle sum proof v")
		return
	}

	log.WithFields(log.Fields{
		"key":    merkleSumTreeVal,
		"length": len(valData),
	}).Info("diskdb load merkle sum proof val")

	vals := bytes.Split(valData, []byte(":"))
	bigVals := make([]*big.Int, len(vals))
	for i := range vals {
		bigVals[i] = big.NewInt(0).SetBytes(vals[i])
	}

	mstTree := tree.NewMerkleSumTree(datas, bigVals)
	w.Lock()
	w.mstTreeCache[date] = mstTree
	w.Unlock()
	log.WithFields(log.Fields{"date": date, "root": mstTree.GetRoot()}).Info("load merkle sum tree")
	return true
}
