package core

import (
	"fmt"
	"math/big"
	"math/rand"
	"sync"
	"time"
	"witness/conf"
	"witness/quest"
	"witness/tree"
	"witness/util"

	"github.com/ethereum/go-ethereum/common"
	"github.com/ethereum/go-ethereum/ethdb/leveldb"
	"github.com/ethereum/go-ethereum/trie"
	witnessv1 "github.com/odysseus/odysseus-protocol/gen/proto/go/witness/v1"
	log "github.com/sirupsen/logrus"
)

type Witness struct {
	lvdb               *leveldb.Database
	db                 *trie.Database
	state              *StateDB
	pendingProof       map[int64]map[common.Address]*witnessv1.ValidatedProof
	pendingWorkload    uint64
	mtTreeCache        map[string]*tree.MerkleTree
	mstTreeCache       map[string]*tree.MerkleSumTree
	date               string
	rpc                *ChainRPC
	containerAddresses []common.Address
	nmAddresses        []common.Address
	q                  *quest.Quest

	cfg *conf.Config

	sync.Mutex
}

func getDBLastDayRoot(diskDB *leveldb.Database) (lastDay string, stateRoot common.Hash) {
	ok, err := diskDB.Has([]byte("lastday"))
	if err != nil {
		panic(err)
	}
	if ok {
		val, _ := diskDB.Get([]byte("lastday"))
		log.WithField("last day", string(val)).Info("diskdb get last day")
		lastDay = string(val)
		root, _ := diskDB.Get([]byte(fmt.Sprintf("sroot:%s", string(val))))
		stateRoot = common.BytesToHash(root)
		log.WithFields(log.Fields{
			"last day":   lastDay,
			"state root": stateRoot.Hex(),
		}).Info("load state from diskdb")
	}
	return lastDay, stateRoot
}

func RunWitness(q *quest.Quest, cfg *conf.Config) *Witness {
	_rpc := newChain(cfg.ChainRPC, cfg.PrivateKey, cfg.StoreContract, cfg.WitnessContract)
	diskDB, err := leveldb.New(fmt.Sprintf("%s/db", cfg.DataDir), 128, 1024, "", false)
	if err != nil {
		panic(err)
	}

	lastDay, stateRoot := getDBLastDayRoot(diskDB)
	stateDB, err := NewStateDB(diskDB, stateRoot)
	if err != nil {
		panic(err)
	}
	w := &Witness{
		lvdb:         diskDB,
		state:        stateDB,
		pendingProof: make(map[int64]map[common.Address]*witnessv1.ValidatedProof),
		mtTreeCache:  make(map[string]*tree.MerkleTree),
		mstTreeCache: make(map[string]*tree.MerkleSumTree),
		rpc:          _rpc,
		date:         lastDay,
		cfg:          cfg,
		q:            q,
	}
	w.LoadMerkleTree(lastDay)
	w.LoadMerkleSumTree(lastDay)
	go w.UpdateContractAddressJob()
	go w.UpdateGlobalWorkloadJob()
	<-time.After(time.Second * 3)
	go w.Ticker()
	// go w.Mock()
	// go w.ProcessDay()

	return w
}

func (w *Witness) AddPendingProof(miner common.Address, proof *witnessv1.ValidatedProof) {
	w.Lock()
	defer w.Unlock()

	var userTodayWorkload uint64
	currentPendingProof, ok := w.pendingProof[w.todayTimestamp()][miner]
	if ok {
		userTodayWorkload = currentPendingProof.Workload
	}

	if w.pendingProof[w.todayTimestamp()] == nil {
		w.pendingProof[w.todayTimestamp()] = make(map[common.Address]*witnessv1.ValidatedProof)
	}
	w.pendingProof[w.todayTimestamp()][miner] = &witnessv1.ValidatedProof{Workload: userTodayWorkload + proof.Workload}

	log.WithFields(log.Fields{
		"miner":            miner.Hex(),
		"current_workload": userTodayWorkload + proof.Workload,
		"workload":         proof.Workload,
		"global_workload":  w.pendingWorkload,
	}).Debug("add validated proof")
}

func (w *Witness) ProcessDayJob() {
	log.Debugln("process day job", w.yesterdayTimestamp(), w.todayTimestamp())
	w.LoadPendingProofs(w.yesterdayTimestamp(), w.todayTimestamp())
	w.date = w.yesterdayString()
	dayProof := w.Commit()
	mstRoot, _, err := w.CommitMST(dayProof)
	if err != nil {
		log.WithError(err).Error("failed to commit merkle sum tree")
		return
	}
	objects := w.state.IterAllObject()
	mtRoot, err := w.CommitMT(objects)
	if err != nil {
		log.WithError(err).Error("failed to commit merkle tree")
		return
	}
	err = w.lvdb.Put([]byte("lastday"), []byte(w.date))
	if err != nil {
		log.WithError(err).Error("db failed to update last day")
		return
	}
	txHash, err := w.rpc.SubmitProofs(w.date, mstRoot, mtRoot)
	if err != nil {
		log.WithError(err).Error("submit proofs")
		return
	}
	err = w.lvdb.Put([]byte(fmt.Sprintf("txid:%s", w.date)), txHash.Bytes())
	if err != nil {
		log.WithError(err).Error("db failed to save txid")
		return
	}
	log.WithField("tx_hash", txHash.Hex()).Info("submit proofs")
	log.Info("process day done")
}

func (w *Witness) Commit() (dayProofs map[common.Address]*witnessv1.ValidatedProof) {
	st := time.Now()
	proof, totalWorkload := w.RefreshPendingProof()
	balancePerWorkload, err := w.rpc.GetWorkloadThreshold(totalWorkload)
	if err != nil {
		log.WithError(err).Error("failed to get workload threshold")
		return
	}

	for miner, proof := range proof {
		balance := big.NewInt(0).Mul(balancePerWorkload, big.NewInt(int64(proof.Workload)))
		proof.Balance = balance.String()
		err := w.SealProof(miner, proof)
		if err != nil {
			log.WithError(err).Error("failed to seal proof")
			return
		}
	}
	root, err := w.state.Commit()
	if err != nil {
		return
	}
	log.WithFields(log.Fields{"root": root.Hex(), "cost": time.Since(st).String()}).Info("commit state root")
	err = w.lvdb.Put([]byte(fmt.Sprintf("sroot:%s", w.date)), root.Bytes())
	if err != nil {
		log.WithError(err).Error("failed to update state root")
		return
	}
	return proof
}

func (w *Witness) SealProof(miner common.Address, proof *witnessv1.ValidatedProof) (err error) {
	minerObject := w.state.GetMinerObject(miner)
	oldBalance := minerObject.Balance
	minerObject.Balance = util.MustAddStringToBigInt(oldBalance, proof.Balance).String()
	err = w.state.UpdateMinerObject(miner, minerObject)
	if err != nil {
		return err
	}
	log.WithFields(log.Fields{
		"miner":       miner.Hex(),
		"old_balance": oldBalance,
		"new_balance": minerObject.Balance,
	}).Info("seal proof")
	return err
}

func (w *Witness) RefreshPendingProof() (proof map[common.Address]*witnessv1.ValidatedProof, totalWorkload uint64) {
	w.Lock()
	defer w.Unlock()
	proof = make(map[common.Address]*witnessv1.ValidatedProof)
	for timestamp, pendingProofMap := range w.pendingProof {
		if timestamp != w.todayTimestamp() {
			continue
		}
		for miner, pendingProof := range pendingProofMap {
			proof[miner] = pendingProof
			totalWorkload += pendingProof.Workload
		}
	}
	w.pendingProof[w.yesterdayTimestamp()] = make(map[common.Address]*witnessv1.ValidatedProof)
	return
}

func (w *Witness) Ticker() {
	executionTime := time.Date(time.Now().Year(), time.Now().Month(), time.Now().Day(), 0, 0, w.cfg.CommitTime, 0, time.UTC).Add(w.duration())
	// executionTime := time.Now().Add(w.duration())
	waitTime := executionTime.Sub(time.Now())
	timer := time.NewTimer(waitTime)
	log.WithField("wait_time", waitTime.String()).Info("prepare commit task")
	for {
		<-timer.C
		w.ProcessDayJob()
		executionTime = executionTime.Add(w.duration())
		waitTime = executionTime.Sub(time.Now())
		timer.Reset(waitTime)
		log.WithField("wait time", waitTime.String()).Info("prepare commit task")
	}
}

func (w *Witness) Mock() {
	addresses := make([]common.Address, 10)
	rand.Seed(time.Now().Unix())
	for i := 0; i < 10; i++ {
		b20 := make([]byte, 20)
		rand.Read(b20)
		addresses[i] = common.BytesToAddress(b20)
	}

	for {
		time.Sleep(time.Second)
		w.AddPendingProof(addresses[rand.Intn(10)], &witnessv1.ValidatedProof{
			Workload:  uint64((rand.Int63n(9) + 1) * 100),
			Timestamp: uint64(time.Now().Unix() - 1),
		})
	}
}

func (w *Witness) IsContainerAddress(address common.Address) bool {
	for _, addr := range w.containerAddresses {
		if addr.Hex() == address.Hex() {
			return true
		}
	}
	return false
}

func (w *Witness) IsNMAddress(address common.Address) bool {
	for _, addr := range w.nmAddresses {
		if addr.Hex() == address.Hex() {
			return true
		}
	}
	return false
}

func (w *Witness) todayString() string {
	return time.Unix(w.todayTimestamp(), 0).Format("2006-01-02")
}

func (w *Witness) yesterdayString() string {
	return time.Unix(w.yesterdayTimestamp(), 0).Format("2006-01-02")
}

func (w *Witness) todayTimestamp() int64 {
	now := time.Now().UTC()
	// today := time.Date(now.Year(), now.Month(), now.Day(), 0, 0, 0, 0, time.UTC)
	today := time.Date(now.Year(), now.Month(), 6, 0, 0, 0, 0, time.UTC)
	return today.Unix()
}

func (w *Witness) yesterdayTimestamp() int64 {
	return w.todayTimestamp() - 86400
}

func (w *Witness) dateToTimestamp(date string) int64 {
	t, _ := time.ParseInLocation("2006-01-02", date, time.UTC)
	return t.Unix()
}

func (w *Witness) duration() time.Duration {
	return 3600 * 24 * time.Second
}
